.g-dicom-view
  margin-bottom 15px

  .g-dicom-left
    width 512px
    float left

  .g-dicom-right
    margin-left 522px
    height 512px
    overflow scroll
    border 1px solid #ccc
    border-radius 4px

  .g-dicom-filename
    text-align center
    margin 5px 0
    white-space nowrap
    overflow hidden

  .g-dicom-controls
    .right
      float right

      button
        margin-right 0
        margin-left 5px

    button
      margin-right 5px

    input
      margin 5px 0
